# Copyright 2020 The Microsoft DeepSpeed Team

PDSH_LAUNCHER = 'pdsh'
PDSH_MAX_FAN_OUT = 1024

OPENMPI_LAUNCHER = 'openmpi'
MPICH_LAUNCHER = 'mpich'
SLURM_LAUNCHER = 'slurm'
MVAPICH_LAUNCHER = 'mvapich'
MVAPICH_TMP_HOSTFILE = '/tmp/deepspeed_mvapich_hostfile'

ELASTIC_TRAINING_ID_DEFAULT = "123456789"
